#include <stack>
#include "common.h"
#include "SceneGraph.h"
#include "X3dExporter.h"
#include "Mesh.h"
#include "KdTree.h"
#include "ViewCellBsp.h"
#include "ViewCell.h"
#include "Polygon3.h"

X3dExporter::X3dExporter(const string filename):Exporter(filename)
{
  stream.open(mFilename.c_str());
  stream<<"<?xml version=\"1.0\" encoding=\"UTF-8\"?>"<<endl;
  stream<<"<X3D>"<<endl;
  stream<<"<Scene>"<<endl;
  
}

X3dExporter::~X3dExporter()
{
  stream<<"</Scene>"<<endl;
  stream<<"</X3D>"<<endl;
  stream.close();
}


bool
X3dExporter::ExportRays(const vector<Ray> &rays,
			const float length,
			const RgbColor &color)
{
  vector<Ray>::const_iterator ri = rays.begin();
  stream<<"<Shape>"<<endl;
  stream<<"<Appearance>"<<endl;
  stream<<"<Material ambientColor=\""<<color.r<<" "<<color.g<<" "<<color.b<<
    "\" />"<<endl;
  stream<<"</Appearance>"<<endl;
  
  stream<<"<IndexedLineSet coordIndex=\""<<endl;

  int index = 0;
  for (; ri != rays.end(); ri++) {
    stream<<index<<" "<<index+1<<" -1\n";
    index+=2;
  }
  
  stream<<"\" >"<<endl;
  
  stream<<"<Coordinate  point=\""<<endl;
  
  ri = rays.begin();
  for (; ri != rays.end(); ri++) {
    Vector3 a = (*ri).GetLoc();
    
    Vector3 b;
    if (length < 0)
      b = (*ri).GetLoc() - length*(*ri).GetDir();
    else
      if ((*ri).intersections.size()==0)
	b = (*ri).GetLoc() + length*(*ri).GetDir();
      else
	b = (*ri).Extrap((*ri).intersections[0].mT);
    
    stream<<a.x<<" "<<a.y<<" "<<a.z<<" ,";
    stream<<b.x<<" "<<b.y<<" "<<b.z<<" ,\n";
  }
  
  stream<<"\" >"<<endl;
  stream<<"</Coordinate>"<<endl;
  stream<<"</IndexedLineSet>"<<endl;
  stream<<"</Shape>"<<endl;
  return true;
}

bool
X3dExporter::ExportRays(const RayContainer &rays,
						const float length,
						const RgbColor &color)
{
  RayContainer::const_iterator ri = rays.begin();

  stream<<"<Shape>"<<endl;
  stream<<"<Appearance>"<<endl;
  stream<<"<Material ambientColor=\""<<color.r<<" "<<color.g<<" "<<color.b<<
    "\" />"<<endl;
  stream<<"</Appearance>"<<endl;
  
  stream<<"<IndexedLineSet coordIndex=\""<<endl;

  int index = 0;
  for (; ri != rays.end(); ri++) {
    stream<<index<<" "<<index+1<<" -1\n";
    index+=2;
  }
  
  stream<<"\" >"<<endl;
  
  stream<<"<Coordinate  point=\""<<endl;
  
  ri = rays.begin();
  for (; ri != rays.end(); ri++) {
    Vector3 a = (*ri)->GetLoc();
    
    Vector3 b;
    if (length < 0)
      b = (*ri)->GetLoc() - length*(*ri)->GetDir();
    else
      if ((*ri)->intersections.size()==0)
	b = (*ri)->GetLoc() + length*(*ri)->GetDir();
      else
	b = (*ri)->Extrap((*ri)->intersections[0].mT);
    
    stream<<a.x<<" "<<a.y<<" "<<a.z<<" ,";
    stream<<b.x<<" "<<b.y<<" "<<b.z<<" ,\n";
  }
  
  stream<<"\" >"<<endl;
  stream<<"</Coordinate>"<<endl;
  stream<<"</IndexedLineSet>"<<endl;
  stream<<"</Shape>"<<endl;

  return true;
}

void
X3dExporter::ExportSceneNode(SceneGraphNode *node)
{
  stream<<"<Group>"<<endl;

  SceneGraphNodeContainer::iterator ni = node->mChildren.begin();
  for (; ni != node->mChildren.end(); ni++)
    ExportSceneNode(*ni);
  
  
  ObjectContainer::const_iterator mi = node->mGeometry.begin();
  for (; mi != node->mGeometry.end(); mi++) {
    // export the transform...
    ExportIntersectable(*mi);
  }
  
  stream<<"</Group>"<<endl;

}
void
X3dExporter::ExportIntersectable(Intersectable *object)
{
  switch (object->Type()) {
  case Intersectable::MESH_INSTANCE:
  case Intersectable::TRANSFORMED_MESH_INSTANCE:
    ExportMeshInstance((MeshInstance *)object);
	break;
  case Intersectable::VIEW_CELL:
	ExportViewCell((ViewCell *)object);
    break;
  default:
    cerr<<"Sorry the export for object not yet defined"<<endl;
    break;
  }
}

void
X3dExporter::ExportMeshInstance(MeshInstance *object)
{
  // $$JB$$
  // in the future check whether the mesh was not already exported
  // and use a reference to the that mesh instead
  ExportMesh(object->GetMesh());
}

void 
X3dExporter::ExportViewCells(const ViewCellContainer &viewCells)
{
	ViewCellContainer::const_iterator it, it_end = viewCells.end();

	for (it = viewCells.begin(); it != it_end; ++ it)
		ExportViewCell(*it);
}

void 
X3dExporter::ExportBspViewCellPartition(const BspTree &tree, const int maxPvs)
{
	ViewCellContainer viewCells;
	tree.CollectViewCells(viewCells);

	ViewCellContainer::const_iterator it, it_end = viewCells.end();


	if (maxPvs == 0)
		mUseForcedMaterial = true;

	for (it = viewCells.begin(); it != it_end; ++ it)
	{
		mForcedMaterial.mDiffuseColor.b = 1.0f;
  
		float importance = 0;

		if (maxPvs > 0)
			importance = (float)(*it)->GetPvs().GetSize() / (float)maxPvs;

		mForcedMaterial.mDiffuseColor.r = importance;
		mForcedMaterial.mDiffuseColor.g = 1.0f - mForcedMaterial.mDiffuseColor.r;
		
		if ((*it)->GetMesh())
			ExportViewCell(*it);
		else
		{
			PolygonContainer cell;
			tree.ConstructGeometry(dynamic_cast<BspViewCell *>(*it), cell);

			ExportPolygons(cell);
		}
	}
}

void 
X3dExporter::ExportBspLeaves(const BspTree &tree)
{
	vector<BspLeaf *> leaves;
	tree.CollectLeaves(leaves);

	vector<BspLeaf *>::const_iterator it, it_end = leaves.end();

	for (it = leaves.begin(); it != it_end; ++ it)
	{
		PolygonContainer cell;
		tree.ConstructGeometry(*it, cell);
		ExportPolygons(cell);
	}
}

void
X3dExporter::ExportViewCell(ViewCell *viewCell)
{
	if (viewCell->GetMesh())
		ExportMesh(viewCell->GetMesh());
}

void
X3dExporter::ExportMesh(Mesh *mesh)
{

  stream<<"<Shape>"<<endl;
  stream<<"<Appearance>"<<endl;
  
  // $$ tmp -> random material
  
  float r, g, b;

  if (mUseForcedMaterial) {
    r = mForcedMaterial.mDiffuseColor.r;
    g = mForcedMaterial.mDiffuseColor.g;
    b = mForcedMaterial.mDiffuseColor.b;
    
  } else
    if (mesh->mMaterial) {
      r = mesh->mMaterial->mDiffuseColor.r;
      g = mesh->mMaterial->mDiffuseColor.g;
      b = mesh->mMaterial->mDiffuseColor.b;
    } else {
      r = RandomValue(0.5, 1.0);
      g = RandomValue(0.5, 1.0);
      b = RandomValue(0.5, 1.0);
    }
  stream<<"<Material diffuseColor=\""<<r<<" "<<g<<" "<<b<<
    "\" specularColor=\"0.0 0.0 0.0\"/>"<<endl;
  stream<<"</Appearance>"<<endl;


  if (mWireframe)
    stream<<"<IndexedLineSet ccw=\"TRUE\" coordIndex=\""<<endl;
  else
    stream<<"<IndexedFaceSet ccw=\"TRUE\" coordIndex=\""<<endl;

  FaceContainer::const_iterator fi = mesh->mFaces.begin();

  int index = 0;
  
  for (; fi != mesh->mFaces.end(); fi++) {
    Face *face = *fi;
    VertexIndexContainer::const_iterator vi = face->mVertexIndices.begin();
    for (; vi != face->mVertexIndices.end(); vi++)
      stream<<*vi<<" ";
	if (mWireframe) // final line to finish polygon
		stream << (*face->mVertexIndices.begin()) << " ";

    stream<<"-1"<<endl;
  }
  stream<<"\" >"<<endl;

  stream<<"<Coordinate  point=\""<<endl;
  
  VertexContainer::const_iterator vi = mesh->mVertices.begin();
  for (; vi != mesh->mVertices.end(); vi++) {
    stream<<(*vi).x<<" "<<(*vi).y<<" "<<(*vi).z;
    stream<<","<<endl;
  }
  
  stream<<"\" >"<<endl;
  stream<<"</Coordinate>"<<endl;

  if (mWireframe)
    stream<<"</IndexedLineSet>"<<endl;
  else
    stream<<"</IndexedFaceSet>"<<endl;
  
  stream<<"</Shape>"<<endl;

}


void X3dExporter::ExportPolygon(Polygon3 *poly)
{
	stream << "<Shape>" << endl;
	stream << "<Appearance>" << endl;
  
	// $$ tmp -> random material
  
	float r, g, b;

	if (mUseForcedMaterial) 
	{
		r = mForcedMaterial.mDiffuseColor.r;
		g = mForcedMaterial.mDiffuseColor.g;
		b = mForcedMaterial.mDiffuseColor.b;
	} 
	else if (poly->mMaterial) 
	{
		r = poly->mMaterial->mDiffuseColor.r;
		g = poly->mMaterial->mDiffuseColor.g;
		b = poly->mMaterial->mDiffuseColor.b;
	} else 
	{
		r = RandomValue(0.5, 1.0);
		g = RandomValue(0.5, 1.0);
		b = RandomValue(0.5, 1.0);
	}

	stream << "<Material diffuseColor=\"" << r << " " << g << " " << b 
		   << "\" specularColor=\"0.0 0.0 0.0\"/>" << endl;

    stream << "</Appearance>" << endl;


	//-- create and write indices
	if (mWireframe)
		stream << "<IndexedLineSet ccw=\"TRUE\" coordIndex=\"" << endl;
	else
		stream << "<IndexedFaceSet ccw=\"TRUE\" coordIndex=\"" << endl;

	int index = 0;
	
	VertexContainer::const_iterator vi;  
	
	for (index = 0; index < (int)poly->mVertices.size(); ++ index) 
		stream << index << " ";
	
	if (mWireframe) // final line to finish polygon
		stream << "0 ";

	stream << "-1" << endl;
	stream << "\" >" << endl;
	
	stream << "<Coordinate  point=\"" << endl;
  
  	for (vi = poly->mVertices.begin(); vi != poly->mVertices.end(); ++vi) 
	{
		stream << (*vi).x << " " << (*vi).y << " " << (*vi).z;
		stream << "," << endl;
	}
  
	stream << "\" >" << endl;
	stream << "</Coordinate>" << endl;

	if (mWireframe)
		stream << "</IndexedLineSet>" << endl;
	else
		stream << "</IndexedFaceSet>" << endl;
  
	stream << "</Shape>" << endl;
}

void X3dExporter::ExportPolygons(const PolygonContainer &polys)
{
  	stream << "<Shape>" << endl;
	stream << "<Appearance>" << endl;
  
	// $$ tmp -> random material
  
	float r, g, b;

	if (mUseForcedMaterial) 
	{
		r = mForcedMaterial.mDiffuseColor.r;
		g = mForcedMaterial.mDiffuseColor.g;
		b = mForcedMaterial.mDiffuseColor.b;
	} 
	else 
	{
		r = RandomValue(0.5, 1.0);
		g = RandomValue(0.5, 1.0);
		b = RandomValue(0.5, 1.0);
	}

	stream << "<Material diffuseColor=\"" << r << " " << g << " " << b 
		   << "\" specularColor=\"0.0 0.0 0.0\"/>" << endl;

    stream << "</Appearance>" << endl;


	//-- create and write indices
	if (mWireframe)
		stream << "<IndexedLineSet ccw=\"TRUE\" coordIndex=\"" << endl;
	else
		stream << "<IndexedFaceSet ccw=\"TRUE\" coordIndex=\"" << endl;

	int index = 0;
	
	PolygonContainer::const_iterator pit;

    VertexContainer::const_iterator vi;  
	
	for (pit = polys.begin(); pit != polys.end(); ++pit)
	{
		Polygon3 *poly = *pit;
		int startIdx = index;
		for (vi = poly->mVertices.begin(); vi != poly->mVertices.end(); ++vi) 
		{
			stream << index ++ << " ";
		}

		stream << startIdx << " ";// finish line
		stream << "-1" << endl;
	}

	stream << "\" >" << endl;
	
	stream << "<Coordinate  point=\"" << endl;
	for (pit = polys.begin(); pit != polys.end(); ++ pit)
	{
		Polygon3 *poly = *pit;
        for (vi = poly->mVertices.begin(); vi != poly->mVertices.end(); ++vi) 
		{
			stream << (*vi).x << " " << (*vi).y << " " << (*vi).z;
			stream << "," << endl;
		}
	}
	stream << "\" >" << endl;
	stream << "</Coordinate>" << endl;

	if (mWireframe)
		stream << "</IndexedLineSet>" << endl;
	else
		stream << "</IndexedFaceSet>" << endl;
  
	stream << "</Shape>" << endl;
}

bool
X3dExporter::ExportBox(const AxisAlignedBox3 &box)
{
  Mesh *mesh = new Mesh;
  // add 6 vertices of the box
  int index = (int)mesh->mVertices.size();
  for (int i=0; i < 8; i++) {
    Vector3 v;
    box.GetVertex(i, v);
    mesh->mVertices.push_back(v);
  }
  
  mesh->AddFace(new Face(index + 0, index + 1, index + 3, index + 2) );
  mesh->AddFace(new Face(index + 0, index + 2, index + 6, index + 4) );
  mesh->AddFace(new Face(index + 4, index + 6, index + 7, index + 5) );
  
  mesh->AddFace(new Face(index + 3, index + 1, index + 5, index + 7) );
  mesh->AddFace(new Face(index + 0, index + 4, index + 5, index + 1) );
  mesh->AddFace(new Face(index + 2, index + 3, index + 7, index + 6) );
  
  ExportMesh(mesh);
  delete mesh;
  return true;
}

bool
X3dExporter::ExportBspTree(const BspTree &tree)
{
	if (mExportRayDensity) 
	{
		return ExportBspTreeRayDensity(tree);
	}
  
	stack<BspNode *> tStack;

	tStack.push(tree.GetRoot());

	Mesh *mesh = new Mesh;

	AxisAlignedBox3 box = tree.GetBoundingBox();
	bool savedWireframe = mWireframe;

	SetWireframe();
	ExportBox(box);
	
	if (!savedWireframe)
		SetFilled();

	//ViewCellContainer foundViewCells;

	if (BspTree::sStoreSplitPolys)
	{
		while (!tStack.empty()) 
		{
            BspNode *node = tStack.top();
    
			tStack.pop();
	
			PolygonContainer::const_iterator it;
			PolygonContainer::const_iterator it_end = node->GetPolygons()->end();

			for (it = node->GetPolygons()->begin(); it != it_end; ++ it)
				ExportPolygon(*it);
			
			if (!node->IsLeaf()) 
			{
				BspInterior *interior = dynamic_cast<BspInterior *>(node);
      
				tStack.push(interior->GetFront());
				tStack.push(interior->GetBack());

			}
		}
	}
	// export view cells
	ExportBspViewCellPartition(tree);	

	return true;
}

bool X3dExporter::ExportKdTree(const KdTree &tree)
{
	 if (mExportRayDensity) {
    return ExportKdTreeRayDensity(tree);
  }
  
  stack<KdNode *> tStack;

  tStack.push(tree.GetRoot());

  Mesh *mesh = new Mesh;
  
  while (!tStack.empty()) {
    KdNode *node = tStack.top();
    tStack.pop();
    AxisAlignedBox3 box = tree.GetBox(node);
    // add 6 vertices of the box
    int index = (int)mesh->mVertices.size();
    for (int i=0; i < 8; i++) {
      Vector3 v;
      box.GetVertex(i, v);
      mesh->mVertices.push_back(v);
    }
    mesh->AddFace(new Face(index + 0, index + 1, index + 3, index + 2) );
    mesh->AddFace(new Face(index + 0, index + 2, index + 6, index + 4) );
    mesh->AddFace(new Face(index + 4, index + 6, index + 7, index + 5) );

    mesh->AddFace(new Face(index + 3, index + 1, index + 5, index + 7) );
    mesh->AddFace(new Face(index + 0, index + 4, index + 5, index + 1) );
    mesh->AddFace(new Face(index + 2, index + 3, index + 7, index + 6) );

    if (!node->IsLeaf()) {
      KdInterior *interior = (KdInterior *)node;
      tStack.push(interior->mFront);
      tStack.push(interior->mBack);
    }
  }
  
  ExportMesh(mesh);
  delete mesh;
  return true;
	// TODO
	return true;
}


bool
X3dExporter::ExportBspTreeRayDensity(const BspTree &tree)
{
	stack<BspNode *> tStack;

	tStack.push(tree.GetRoot());

	bool fm = mUseForcedMaterial;
	
	mUseForcedMaterial = true;
	
	mForcedMaterial.mDiffuseColor.g = 1.0f;
  	mForcedMaterial.mDiffuseColor.b = 1.0f;
  
	while (!tStack.empty()) 
	{
		BspNode *node = tStack.top();
		tStack.pop();

		if (node->IsLeaf()) 
		{
			ViewCell *vc = dynamic_cast<BspLeaf *>(node)->GetViewCell();
     
			// set the mesh material according to the ray density
			if (vc->mPassingRays.mRays) 
			{
				float importance = 
					vc->mPassingRays.mContributions / (float)vc->mPassingRays.mRays;

				mForcedMaterial.mDiffuseColor.r = importance;
				mForcedMaterial.mDiffuseColor.g = 1.0f - mForcedMaterial.mDiffuseColor.r;
				ExportViewCell(vc);
			}  
		} else 
		{
			BspInterior *interior = (BspInterior *)node;
			tStack.push(interior->GetFront());
			tStack.push(interior->GetBack());
		}
	}
  
	// restore the state of forced material
	mUseForcedMaterial = fm;

	return true;
}

bool
X3dExporter::ExportKdTreeRayDensity(const KdTree &tree)
{
  stack<KdNode *> tStack;

  tStack.push(tree.GetRoot());

  bool fm = mUseForcedMaterial;
  mUseForcedMaterial = true;
  mForcedMaterial.mDiffuseColor.g = 1.0f;
  mForcedMaterial.mDiffuseColor.b = 1.0f;
  while (!tStack.empty()) {
    KdNode *node = tStack.top();
    tStack.pop();
    if (node->IsLeaf()) {
      AxisAlignedBox3 box = tree.GetBox(node);
      Mesh *mesh = new Mesh;
      
      // add 6 vertices of the box
      int index = (int)mesh->mVertices.size();
      for (int i=0; i < 8; i++) {
	Vector3 v;
	box.GetVertex(i, v);
	mesh->mVertices.push_back(v);
      }
      mesh->AddFace(new Face(index + 0, index + 1, index + 3, index + 2) );
      mesh->AddFace(new Face(index + 0, index + 2, index + 6, index + 4) );
      mesh->AddFace(new Face(index + 4, index + 6, index + 7, index + 5) );
      
      mesh->AddFace(new Face(index + 3, index + 1, index + 5, index + 7) );
      mesh->AddFace(new Face(index + 0, index + 4, index + 5, index + 1) );
      mesh->AddFace(new Face(index + 2, index + 3, index + 7, index + 6) );


      // set the mesh material according to the ray density
      KdLeaf *leaf = (KdLeaf *) node;
      if (leaf->mPassingRays.mRays) {
	float importance = leaf->mPassingRays.mContributions/(float)leaf->mPassingRays.mRays;
	//	float importance = leaf->mPassingRays.mContributions/1000.0f;
	//	float importance = leaf->mPassingRays.mRays/1000.0f;
	///(float)leaf->mPassingRays.mRays;
	// mForcedMaterial.mDiffuseColor.r = log10(leaf->mPassingRays.mRays)/3.0f;
	mForcedMaterial.mDiffuseColor.r = importance;
	mForcedMaterial.mDiffuseColor.g = 1.0f - mForcedMaterial.mDiffuseColor.r;
	ExportMesh(mesh);
      }
      delete mesh;
    } else {
      KdInterior *interior = (KdInterior *)node;
      tStack.push(interior->mFront);
      tStack.push(interior->mBack);
    }
  }
  // restore the state of forced material
  mUseForcedMaterial = fm;
  return true;
}


struct BspSplitData
{
	/// the current node
	BspNode *mNode;

	vector<Plane3 *> mPlanes;
	vector<bool> mSides;
	bool mIsFront;

	BspSplitData(BspNode *node): 
	mNode(node), mIsFront(false)
	{};	
	BspSplitData(BspNode *node, 
				vector<Plane3 *> planes, 
				vector<bool> sides, 
				bool isFront): 
	mNode(node), mPlanes(planes), 
	mSides(sides), mIsFront(isFront)
	{};
};

void X3dExporter::ExportLeavesGeometry(const BspTree &tree, const vector<BspLeaf *> &leaves)
{
	vector<BspLeaf *>::const_iterator it, it_end = leaves.end();

	for (it = leaves.begin(); it != it_end; ++ it)
	{
		PolygonContainer cell;
		tree.ConstructGeometry(*it, cell);
		
		ExportPolygons(cell);

		CLEAR_CONTAINER(cell);
	}
}

void X3dExporter::ExportBspSplits(const BspTree &tree)
{
	std::stack<BspSplitData> tStack;

	BspSplitData tData(tree.GetRoot());
	tStack.push(tData);
  
	PolygonContainer polys;

	while (!tStack.empty())
	{
		// filter polygons donw the tree
		BspSplitData tData = tStack.top();
	    tStack.pop();	
		
		if (!tData.mNode->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(tData.mNode);

			if (tData.mNode != tree.GetRoot())
				tData.mSides.push_back(tData.mIsFront); // add current side of split plane

			// bounded plane is added to the polygons
			Polygon3 *planePoly = tree.GetBoundingBox().CrossSection(*interior->GetPlane());
		
			// do all the splits with the previous planes
			for (int i = 0; i < (int)tData.mPlanes.size(); ++ i)
			{
				VertexContainer splitPts;
				
				if (planePoly->ClassifyPlane(*tData.mPlanes[i]) == Polygon3::SPLIT)
				{
					Polygon3 *frontPoly = new Polygon3();
					Polygon3 *backPoly = new Polygon3();

					planePoly->Split(*tData.mPlanes[i], *frontPoly, *backPoly, splitPts);
					DEL_PTR(planePoly);

					if(tData.mSides[i] == true)
					{
						planePoly = frontPoly;
						DEL_PTR(backPoly);
					}
					else
					{
						planePoly = backPoly;
						DEL_PTR(frontPoly);
					}
				}
			}

			tData.mPlanes.push_back(interior->GetPlane()); // add plane to split planes

			if (planePoly->Valid())
				polys.push_back(planePoly);
			else
				DEL_PTR(planePoly);
			
			// push the children on the stack
			tStack.push(BspSplitData(interior->GetFront(), tData.mPlanes, tData.mSides, true));
			tStack.push(BspSplitData(interior->GetBack(), tData.mPlanes, tData.mSides, false));
		}
	}	
	ExportPolygons(polys);
	CLEAR_CONTAINER(polys);
}

void X3dExporter::ExportBspSplitPlanes(const BspTree &tree)
{
	std::stack<BspNode *> tStack;

	tStack.push(tree.GetRoot());
  
	PolygonContainer polys;

	while (!tStack.empty())
	{
		// filter polygons donw the tree
		BspNode *node = tStack.top();
	    tStack.pop();	
		
		if (!node->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			// bounded plane is added to the polygons
			polys.push_back(tree.GetBoundingBox().CrossSection(*interior->GetPlane()));
		
			// push the children on the stack
			tStack.push(interior->GetBack());
			tStack.push(interior->GetFront());
		}
	}

	ExportPolygons(polys);
	CLEAR_CONTAINER(polys);
}
